use core::fmt;
use std::{path::{PathBuf, Path}, io};

#[derive(Debug)]
pub enum AviaryError {
	/// One of the files passed by the user simply did not exist
	///
	/// The attached path is the file which didn't exist
	///
	/// Handling:  Halt execution before any upload begins and alert the user of any
	/// missing files.
	FileDNE(PathBuf),

	/// The program lacks permission to read one of the images provided by the user
	///
	/// Handling:  Halt execution before any upload begins and alert the user of the issue
	ReadPermissionDenied(PathBuf),

	/// There was an issue reading data from the disk
	///
	/// Handling:  Halt execution immediately and alert the user
	StreamReadError(PathBuf, io::Error),

	/// One or more of the images wasn't correctly encoded
	///
	/// Handling:  Halt execution immediately and ask the user if they're sure the
	/// provided file is an image.
	ImageFormatError(PathBuf),

	/// The server is currently having some difficulties
	///
	/// Represents a 5XX or 4XX error, as well as any transport error that doesn't seem
	/// like it could stem from a network issue.
	///
	/// Handling:  Halt execution and alert the user that the server isn't working at the
	/// minute, and ask them to try a different server or try again.
	ServerError(String),

	/// Couldn't connect to the server
	///
	/// Indicates that there was some network error which prevented the client from
	/// reaching the server.
	///
	/// Handling:  Halt execution and ask the user to check that their computer is
	/// connected to the internet, and is not having any DNS issues.
	ConnectionError(String),

	/// The server URL provided by the user was invalid
	///
	/// Handling:  Halt execution and alert the user that the server url they provided was
	/// invalid, including a couple examples
	BadServerParameter,
}

impl AviaryError {
	pub fn from_open_error(e: io::ErrorKind, location: &Path) -> AviaryError {
		match e {
			io::ErrorKind::NotFound => AviaryError::FileDNE(location.to_owned()),
			io::ErrorKind::PermissionDenied => AviaryError::ReadPermissionDenied(location.to_owned()),
			_ => panic!(
				"Received an error kind that should be impossible for {}: {:?}",
				location.display(),
				e
			)
		}
	}

	pub fn from_upload_error(err: ureq::Error) -> AviaryError {
		match err {
			ureq::Error::Status(code, msg) => AviaryError::ServerError(
				format!("Error code {} received from server: {}", code,
						msg.into_string().unwrap_or(String::new()))),
			ureq::Error::Transport(transport) => match transport.kind() {
				ureq::ErrorKind::InvalidUrl =>
					AviaryError::BadServerParameter,
				ureq::ErrorKind::Dns =>
					AviaryError::ConnectionError(
						format!("DNS issue: {}", transport.message().unwrap_or(""))),
				ureq::ErrorKind::Io =>
					AviaryError::ConnectionError(
						format!("IO issue: {}", transport.message().unwrap_or(""))),
				ureq::ErrorKind::ConnectionFailed =>
					AviaryError::ConnectionError(
						format!("Connection issue: {}", transport.message().unwrap_or(""))),
				ureq::ErrorKind::TooManyRedirects =>
					AviaryError::ServerError("Too many redirects".to_owned()),
				ureq::ErrorKind::BadHeader =>
					AviaryError::ServerError("Invalid header from server".to_owned()),
				ureq::ErrorKind::BadStatus =>
					AviaryError::ServerError("Invalid status from server".to_owned()),
				unk => panic!("Unexpected transport error kind {unk}:\n{transport}")
			},
		}
	}
}

impl fmt::Display for AviaryError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        writeln!(f, "\x1b[31mError!\x1b[0m\n")?;
		match self {
			Self::FileDNE(path) =>
				writeln!(f, "The file at \x1b[36m{}\x1b[0m does not exist", path.display()),

			Self::ReadPermissionDenied(path) =>
				writeln!(f, "Permission denied for the file \x1b[36m{}\x1b[0m", path.display()),

			Self::StreamReadError(path, err) =>
				writeln!(f,
					concat!(
						"Something bad happened while we were reading from ",
						"\x1b[36m{}\x1b[0m.  This could be caused by removing a ",
						"drive/phone without unplugging it, or several other possible ",
						"things.  This error might give you a clue:\n\n{}"
					),
					path.display(), err
				),

			Self::ImageFormatError(path) =>
				writeln!(f,
					concat!(
						"We had a little bit of trouble understanding the image at ",
						"\x1b[36m{}\x1b[0m.  Normally, this indicates that it might not be ",
						"an image file (e.g. if it's a video instead).  If you're sure that ",
						"it's an image, this could indicate that it's corrupted or of an ",
						"unsupported format.\n\n",
						"\x1b[34mSupported formats:\x1b[0m\n",
						" - gif\n",
						" - jpg\n",
						" - ico\n",
						" - png\n",
						" - tiff\n",
						" - webp\n",
						" - bmp\n",
						" - hdr",
					),
					path.display(),
				),

			Self::ServerError(msg) =>
				writeln!(f,
					concat!(
						"The server seems to be down or misbehaving at the minute.  ",
						"Normally, this issue should resolve by itself once the server ",
						"maintainers identify the problem and fix it.\n\n",
						"If the issue persists, please submit an issue on the bug ",
						"tracker.  This error message may help to identify the problem:",
						"\n\n\x1b[31m{}\x1b[0m",
					),
					msg,
				),

			Self::ConnectionError(msg) =>
				writeln!(f,
					concat!(
						"There was an issue connecting to the server!  Check that your ",
						"internet is online, and you can access websites normally.  ",
						"This information might help diagnose the issue:\n\n",
						"\x1b[31m{}\x1b[0m",
					),
					msg,
				),

			Self::BadServerParameter =>
				writeln!(f,
					concat!(
						"The server URL you provided was \x1b[31minvalid\x1b[0m!  ",
						"Please check to make sure that it is properly formatted, like ",
						"\x1b[1m0x0.st\x1b[0m or \x1b[1menvs.sh\x1b[0m.",
					),
				),
		}
    }
}
